// $Id: admin.cc,v 1.2 1999/05/04 19:18:19 mencnar Exp $
//
// $Source: /home/iris/mencnar/ec/CVS/src/cc/admin.cc,v $
//
// $Log: admin.cc,v $
// Revision 1.2  1999/05/04 19:18:19  mencnar
// Added support for creating channels with desired QoS parameters.
// Fixes in ORBIX port.
//
// Revision 1.1  1999/02/24 13:01:53  mencnar
// *** empty log message ***
//
//

// Contains implementation of ECAdmin::Factory interface
// responsible for creating/destroying event channels
//

#ifdef ORBIX
#define EXCEPTIONS
#endif

#include "ec.hh"
#include "admin.hh"
#include "bootstrap.hh"
#include <stdio.h>
#include "trace.hh"

ECFactory_i::
ECFactory_i(
#ifndef NO_MCAST
	    McastSourceFactory *_mcast_src_factory,
	    CosNaming::NamingContext_ptr nsroot,
	    const char *context_name,
	    unsigned long slave_keep_alive
#endif
	    ) 
#ifndef NO_MCAST
  : mcast_src_factory( _mcast_src_factory)
#endif
{
  CONNECT_IMPL(Bootstrap::boa_ptr, this);
#ifndef NO_MCAST
  mcast_master= new McastMaster_i( *mcast_src_factory,
				   nsroot,
				   context_name,
				   slave_keep_alive);
#else
  mcast_master= NULL;
#endif
}

xstring ECFactory_i::
get_unique_name( const MapEChannels &ecs) {
  // PRECONDITION:
  // a read (or write) lock is held
  //
  
  char buf[100];
  unsigned long id= 0;
  do sprintf(buf,"UQEC%ul", id++);
  while ( ecs.find( xstring( (const char*) buf)) != ecs.end());
  return xstring( (const char *) buf);
}

void ECFactory_i::
channel_disposed( EventChannel_i *eci) {
  GeneralLock::auto_write gawl( guard);
  MapEChannels::iterator it,it2;
  for (it= echannels.begin();
       it != echannels.end();) {
    it2= it;
    it++;
    if ((*it2).second == eci) echannels.erase(it2);
  }
  for (it= techannels.begin();
       it != techannels.end();) {
    it2= it;
    it++;
    if ((*it2).second == eci) techannels.erase(it2);
  }
}

ECAdmin::ChannelDescriptions * ECFactory_i::
channels( ENV_DEF1(ev)) {
  GeneralLock::auto_read garl( guard);
  ECAdmin::ChannelDescriptions *ret=
    new ECAdmin::ChannelDescriptions( echannels.size());
  ret->length( echannels.size());
  CORBA::ULong i= 0;
  for (MapEChannels::iterator it= echannels.begin();
       it != echannels.end();
       it++, i++) {
    (*ret)[i].name= (const char *) (*it).first.c_str();
    (*ret)[i].channel= CosEventChannelAdmin::EventChannel::
      _duplicate( (*it).second->corba_object);
    (*ret)[i].max_queue_size= (*it).second->get_max_queue();
    (*ret)[i].overflow_action= (*it).second->get_overflow_action();
    (*ret)[i].not_connected_timeout= (*it).second->not_connected_timeout;
    (*ret)[i].keep_alive_time= (*it).second->keep_alive_time;
  }
  return ret;
}

ECAdmin::TypedChannelDescriptions * ECFactory_i::
typed_channels( ENV_DEF1(ev)) {

  GeneralLock::auto_read garl( guard);
  ECAdmin::TypedChannelDescriptions *ret=
    new ECAdmin::TypedChannelDescriptions( techannels.size());
  ret->length( techannels.size());
  CORBA::ULong i= 0;
  for (MapEChannels::iterator it= techannels.begin();
       it != techannels.end();
       it++, i++) {
    (*ret)[i].name= (const char *) (*it).first.c_str();
    (*ret)[i].channel= CosTypedEventChannelAdmin::TypedEventChannel::
      _duplicate( (*it).second->typed_corba_object);
    (*ret)[i].idef= IRMOD InterfaceDef::
      _duplicate( (*it).second->idef);
    (*ret)[i].max_queue_size= (*it).second->get_max_queue();
    (*ret)[i].overflow_action= (*it).second->get_overflow_action();
    (*ret)[i].not_connected_timeout= (*it).second->not_connected_timeout;
    (*ret)[i].keep_alive_time= (*it).second->keep_alive_time;
  }
  return ret;
}


CosEventChannelAdmin::EventChannel_ptr ECFactory_i::
create_event_channel( const char *name,
		      CORBA::ULong max_queue_size,
		      ECAdmin::OverflowAction oa,
		      CORBA::ULong not_connected_timeout,
		      CORBA::ULong keep_alive_time
		      ENV_DEF(ev)) {

  GeneralLock::auto_write gawl( guard);
  if (echannels.find( xstring(name)) == echannels.end()) {
    trace.start(10) << "create_event_channel ....." << '\n';
    EventChannel_i *impl= 
      new EventChannel_i( max_queue_size, // maximum event queue size
			  oa,
			  not_connected_timeout,
			  keep_alive_time,
			  (mcast_master ? mcast_master->create_forwarder() :
			   AbstractForwarderInstance(NULL,0)),
			  this);

    echannels[xstring(name)]= impl;
    trace.start(10) << "create_event channel ..done" << '\n';
    return CosEventChannelAdmin::EventChannel::
      _duplicate( impl->corba_object);
  }
  // else throw an exception
  else {
    trace.start(10) << "duplicate name" << '\n';
    return NULL;
  }
}
  
CosEventChannelAdmin::EventChannel_ptr ECFactory_i::
create_unique_event_channel( char * &name,
			     CORBA::ULong max_queue_size,
			     ECAdmin::OverflowAction oa,
			     CORBA::ULong not_connected_timeout,
			     CORBA::ULong keep_alive_time
			     ENV_DEF(ev)) {
  GeneralLock::auto_write gawl( guard);
  xstring s( get_unique_name(echannels));
  CosEventChannelAdmin::EventChannel_ptr ret=
    create_event_channel( s.c_str(),
			  max_queue_size,
			  oa,
			  not_connected_timeout,
			  keep_alive_time);
  name= CORBA::string_dup( s.c_str());
  return ret;
}

CosTypedEventChannelAdmin::TypedEventChannel_ptr ECFactory_i::
create_typed_event_channel( IRMOD::InterfaceDef_ptr idef,
			    const char *name,
			    CORBA::ULong max_queue_size,
			    ECAdmin::OverflowAction oa,
			    CORBA::ULong not_connected_timeout,
			    CORBA::ULong keep_alive_time
			    ENV_DEF(ev)) {
#ifdef ORBIX
  throw CORBA::NO_IMPLEMENT(0,CORBA::COMPLETED_NO);
#endif
  GeneralLock::auto_write gawl( guard);
  if (techannels.find( xstring(name)) == techannels.end()) {
    EventChannel_i *impl= 
      new EventChannel_i( max_queue_size, // maximum event queue size
			  oa,
			  not_connected_timeout,
			  keep_alive_time,
			  (mcast_master ? mcast_master->create_forwarder() :
			   AbstractForwarderInstance(NULL,0)),
			  idef,
			  this);

    techannels[xstring(name)]= impl;
    return CosTypedEventChannelAdmin::TypedEventChannel::
      _duplicate( impl->typed_corba_object);
  }
  // else throw an exception
  else return NULL;
}
  
CosTypedEventChannelAdmin::TypedEventChannel_ptr ECFactory_i::
create_unique_typed_event_channel( IRMOD::InterfaceDef_ptr idef,
				   char * &name,
				   CORBA::ULong max_queue_size,
				   ECAdmin::OverflowAction oa,
				   CORBA::ULong not_connected_timeout,
				   CORBA::ULong keep_alive_time
				   ENV_DEF(ev)) {
  GeneralLock::auto_write gawl( guard);
  xstring s= get_unique_name(techannels);
  CosTypedEventChannelAdmin::TypedEventChannel_ptr ret=
    create_typed_event_channel( idef, s.c_str(),
				max_queue_size,
				oa,
				not_connected_timeout,
				keep_alive_time);
  name= CORBA::string_dup( s.c_str());
  return ret;
}
  
CosEventChannelAdmin::EventChannel_ptr ECFactory_i::
find_event_channel( const char *name
		    ENV_DEF(ev)) {
  GeneralLock::auto_read gawl( guard);
  MapEChannels::iterator it;
  if ( (it= echannels.find( xstring(name))) != echannels.end())
    return CosEventChannelAdmin::EventChannel::
      _duplicate( (*it).second->corba_object);
  else return NULL;
}
  
CosTypedEventChannelAdmin::TypedEventChannel_ptr ECFactory_i::
find_typed_event_channel( const char *name
			  ENV_DEF(ev)) {
  GeneralLock::auto_read gawl( guard);
  MapEChannels::iterator it;
  if ( (it= techannels.find( xstring(name))) != techannels.end())
    return CosTypedEventChannelAdmin::TypedEventChannel::
      _duplicate( (*it).second->typed_corba_object);
  else return NULL;
}

void ECFactory_i::
shutdown_factory( ENV_DEF1(ev)) {

  cerr << "Shutdown: not implemented" << endl;
}




