// $Id: forward.cc,v 1.1 1999/02/24 13:01:57 mencnar Exp $
//
// $Source: /home/iris/mencnar/ec/CVS/src/cc/forward.cc,v $
//
// $Log: forward.cc,v $
// Revision 1.1  1999/02/24 13:01:57  mencnar
// *** empty log message ***
//
//

#include <ecmcast.hh>
#include <ecadmin.hh>
#include <iostream.h>
#include <stdlib.h>
#include "compatibility.hh"
#include "bootstrap.hh"
#include "locks.hh"
#include "ec.hh"
#include "mcast.hh"

#include <map>

class McastMaster_i;

class Forwarder_i : public AbstractForwarder {
  omni_mutex guard;
  XSTD::map<CORBA::ULong,ECMcast::Slave_ptr> forwards;
  EventChannel_i *ec_i;
  CORBA::ULong channel_id;
  McastMaster_i &master;
  McastSource &msource;
  CORBA::ULong max_queue_size;
  ECAdmin::OverflowAction overflow_action;
  char *unique_description;
public:
  Forwarder_i( McastMaster_i &_master,
	       McastSource &_msource) :
    master( _master),
    msource( _msource),
    unique_description( _msource.get_unique_description()) {}

  virtual void
  open_event_channel( EventChannel_i &channel,
		      CORBA::ULong _channel_id,
		      CORBA::ULong _max_queue_size,
		      ECAdmin::OverflowAction _oa);

  virtual int
  try_forward_consumer( CosEventChannelAdmin::ProxyPushSupplier_ptr push_supp,
		        CORBA::ULong consumer_id,
			CosEventComm::PushConsumer_ptr push_cons);

  virtual int
  try_forward_typed_consumer( CosEventChannelAdmin::ProxyPushSupplier_ptr push_supp,
			      CORBA::ULong consumer_id,
			      IRMOD InterfaceDef_ptr idef,
			      CosEventComm::PushConsumer_ptr cons_cons);
			   
  virtual void 
  remove_forwarded_consumer( CORBA::ULong consumer_id);
  
  virtual void 
  close_event_channel();
  // called by event channel from operation destroy,
  // AbstractForwarder should destroy all references of this event channel
  //

  virtual void
  push( ECDHandler &ecdh);

  virtual ~Forwarder_i() {
    delete &msource;
    delete []unique_description;
  }
};

class McastMaster_i
  : public BOA_IMPL_BASE(ECMcast,Master) 
{
  XSTD::map<ECMcast::Slave_ptr,ECMcast::SlaveAddresses> slaves;
  XSTD::map<ECMcast::ChannelID,Forwarder_i*> forwarders;
  McastSourceFactory &msource_factory;
  GeneralLock guard;
  
 public:
  McastMaster_i( McastSourceFactory &_msource_factory,
		 CosNaming::NamingContext_ptr root_ns,
		 const char *context_name);

  AbstractForwarder *create_forwarder();

  void insert_forwarder( CORBA::ULong channel_id,
			 Forwarder_i* forw) {
    GeneralLock::auto_write gawl(guard);
    forwarders[channel_id]= forw;
  }
  
  void remove_forwarder( CORBA::ULong channel_id) {
    GeneralLock::auto_write gawl(guard);
    forwarders.erase( channel_id);
  }

  ECMcast::Slave_ptr get_slave( CosEventComm::PushConsumer_ptr push_cons);

  void register_slave( ECMcast::Slave_ptr new_slave,
		       const ECMcast::SlaveAddresses &addresses);
  void consumer_failed( ECMcast::ChannelID channel_id,
			ECMcast::ConsumerID consumer_id);
};

//-------------------------------------------------------------Forwarder_i--
void Forwarder_i::
open_event_channel( EventChannel_i &channel,
		    CORBA::ULong _channel_id,
		    CORBA::ULong _max_queue_size,
		    ECAdmin::OverflowAction _oa) {
  ec_i= &channel;
  channel_id= _channel_id;
  max_queue_size= _max_queue_size;
  overflow_action= _oa;
  // this operation is called only once before any other operation
  master.insert_forwarder( channel_id, this);
}

void Forwarder_i::
close_event_channel() {
  master.remove_forwarder( channel_id);
}

int Forwarder_i::
try_forward_consumer( CosEventChannelAdmin::ProxyPushSupplier_ptr push_supp,
		      CORBA::ULong consumer_id,
		      CosEventComm::PushConsumer_ptr push_cons) {
  ECMcast::Slave_var slave= master.get_slave( push_cons);
  omni_mutex_lock oml(guard);
  if (CORBA::is_nil(slave)) return 0;
  try {
    int ret= (int) slave->forward_push_consumer( unique_description,
						 channel_id,
						 consumer_id,
						 push_supp,
						 push_cons,
						 max_queue_size,
						 overflow_action);
    if (ret) forwards[consumer_id]= ECMcast::Slave::_duplicate(slave);
  }
  catch (...) {
    return 0;
  }
}

int Forwarder_i::
try_forward_typed_consumer( CosEventChannelAdmin::ProxyPushSupplier_ptr push_supp,
			    CORBA::ULong consumer_id,
			    IRMOD InterfaceDef_ptr idef,
			    CosEventComm::PushConsumer_ptr push_cons) {
  ECMcast::Slave_var slave= master.get_slave( push_cons);
  omni_mutex_lock oml(guard);
  if (CORBA::is_nil(slave)) return 0;
  try {
    int ret= (int) slave->forward_typed_push_consumer( unique_description,
						       channel_id,
						       consumer_id,
						       idef,
						       push_supp,
						       push_cons,
						       max_queue_size,
						       overflow_action);
    if (ret) forwards[consumer_id]= ECMcast::Slave::_duplicate(slave);
  }
  catch (...) {
    return 0;
  }
}

void Forwarder_i::
remove_forwarded_consumer( CORBA::ULong consumer_id) {
  omni_mutex_lock oml(guard);
  XSTD::map<CORBA::ULong,ECMcast::Slave_ptr>::iterator it=
    forwards.find( consumer_id);
  if (it != forwards.end()) {
    try {
      (*it).second->close_proxy( channel_id, (*it).first, 1);
    }
    catch (...) {
    }
    CORBA::release( (*it).second);
    forwards.erase( it);
  }
}

void Forwarder_i::
push( ECDHandler &ecdh) {
}

//------------------------------------------------------------McastMaster_i--
McastMaster_i::
McastMaster_i( McastSourceFactory &_msource_factory,
	       CosNaming::NamingContext_ptr root_ns,
	       const char *context_name) :
  msource_factory( _msource_factory)
{

  CONNECT_IMPL(Bootstrap::boa_ptr, this);

  CosNaming::Name ncontext= Bootstrap::create_name(context_name);
  CosNaming::Name nm( ncontext);
  nm.length( nm.length()+1);
  nm[nm.length()-1].id= 
    nm[nm.length()-1].kind= (const char *) "Master";
  try {
    root_ns->unbind( nm);
  }
  catch (...) {}
  try {
    Bootstrap::incremental_bind( root_ns, nm, this);
  }
  catch (CORBA::Exception &ex) {
    cerr << "Could not register ECMcast::Master reference in NS" << endl;
    cerr << ex << endl;
    exit(1);
  }

  // 
  // Now, find all slaves
  //
  CORBA::Object_var ctx_obj= root_ns->resolve(ncontext);
  CosNaming::NamingContext_var ctx=
    CosNaming::NamingContext::_narrow( ctx_obj);
  CosNaming::BindingList_var dir_list;
  CosNaming::BindingIterator_var autoiter;

  ctx->list( 100, dir_list, autoiter);
  do {
    for (CORBA::ULong i=0;
	 i<dir_list->length();
	 i++)
      if ( dir_list[ i].binding_type == CosNaming::nobject) {
	try {
	}
	catch (...) {}
      }
      
    if (CORBA::is_nil(autoiter)) break;
    else if (!autoiter->next_n( 100, dir_list)) break;
  } while (1);
  if (!CORBA::is_nil(autoiter)) autoiter->destroy();
}

AbstractForwarder * McastMaster_i::
create_forwarder() {
  return new Forwarder_i( *this, *(msource_factory.create(NULL,0)));
}

ECMcast::Slave_ptr McastMaster_i::
get_slave( CosEventComm::PushConsumer_ptr push_cons) {
  GeneralLock::auto_read gard(guard);

  // a prototype implementation
  // the real implementation will check the destionation addresss
  // contained in IIOP profile of "push_cons" reference,
  // then find an apropriate slave
  // 
  XSTD::map<ECMcast::Slave_ptr,ECMcast::SlaveAddresses>::iterator it=
    slaves.begin();
  if (it != slaves.end()) 
    return ECMcast::Slave::_duplicate( (*it).first);
  else
    return NULL;
}

void McastMaster_i::
register_slave( ECMcast::Slave_ptr new_slave,
		const ECMcast::SlaveAddresses &addresses) {

  GeneralLock::auto_write gawl(guard);
  XSTD::map<ECMcast::Slave_ptr,ECMcast::SlaveAddresses>::iterator it=
    slaves.find(new_slave);
  slaves[it==slaves.end() ? ECMcast::Slave::_duplicate( new_slave) :
	new_slave]= addresses;
}

void McastMaster_i::
consumer_failed( ECMcast::ChannelID channel_id,
		 ECMcast::ConsumerID consumer_id) {
  
}
