// $Id: pushsupp.cc,v 1.3 1999/05/04 19:34:56 mencnar Exp $
//
// $Source: /home/iris/mencnar/ec/CVS/src/cc/pushsupp.cc,v $
//
// $Log: pushsupp.cc,v $
// Revision 1.3  1999/05/04 19:34:56  mencnar
// Clock synchronization for performance measurements.
//
// Revision 1.2  1999/02/24 12:58:50  mencnar
// *** empty log message ***
//
//

#include "mtime.hh"
#include <iostream.h>
#include <strstream.h>
#include <stdlib.h>
#include <CosEventChannelAdmin.hh>
#include <CosTypedEventChannelAdmin.hh>
#include <TestIntf.hh>
#include <omnithread.h>
#include "bootstrap.hh"
#include "compatibility.hh"

void kill_worker(void*) {
  omni_thread::sleep(1);
  exit(0);
}

class PushSupplier_I : public BOA_IMPL_BASE(CosEventComm,PushSupplier) 
{
  void disconnect_push_supplier( ENV_DECL1(env)) {
    cerr << "PushSupplier disconnected, exitting..." << endl;
    (new omni_thread(kill_worker,NULL))->start();
  }
};

void impl_is_ready( void *) {
  IMPL_IS_READY(Bootstrap::boa_ptr);
}

void usage( const char *prog_name) {
  cerr << prog_name << " event_channel_name_in_NameServer [typed]" << endl;
  exit(1);
}

int main(int argc, char **argv) {
  Bootstrap::init(argc,argv);
  if (argc!=2 && argc!=3) usage(argv[0]);

  CosNaming::NamingContext_var nsroot;
  try {
    nsroot= Bootstrap::getNS();
  }
  catch (CORBA::Exception &ex) {
    cerr << "Could not resolve Naming Service reference " << ex << endl;
  }

  //---- should be deleted in final version, added for performance testing
  
  ECMcast::Master_var master;
  try {
    CosNaming::Name n(2);
    n.length(2);
    n[0].id= (const char *) "a";
    n[0].kind= (const char *) "a";
    n[1].id= (const char *) "Master";
    n[1].kind= (const char *) "Master";
    master= ECMcast::Master::_narrow( nsroot->resolve(n));
  }
  catch (...) {
  }
  
  synch_time( 300, master);
  omni_thread::sleep(90);
  synch_time( 300, master);
  
  
  //-----
  
  if (argc==2) {
    CosEventChannelAdmin::EventChannel_var ec;
    
    try {
      ec= CosEventChannelAdmin::EventChannel::_narrow(
		 nsroot->resolve( Bootstrap::create_name(argv[1])));
    }
    catch (...) {}
    
    if (CORBA::is_nil(ec)) {
      cerr << "Could not find Event Channel" << endl;
      return 1;
    }
    CosEventChannelAdmin::ProxyPushConsumer_var ppushcons;
    try {
      CosEventChannelAdmin::SupplierAdmin_var sa=
	ec->for_suppliers();
      cerr << "for_suppliers" << endl;
      ppushcons= sa->obtain_push_consumer();
    }
    catch (CORBA::SystemException &se) {
      cerr << "SystemException" << se << endl;
    }
    catch (...) {}

    if (CORBA::is_nil(ppushcons)) {
      cerr << "Could not obtain ProxyPushConsumer from EventChannel" << endl;
      return 1;
    }

    PushSupplier_I &pushsupi= *(new PushSupplier_I());
    CONNECT_IMPL(Bootstrap::boa_ptr,&pushsupi);
    try {
      ppushcons->connect_push_supplier( &pushsupi);
    }
    catch (CosEventChannelAdmin::AlreadyConnected &) {
      cerr << "Already connected" << endl;
      return 1;
    }
    catch (...) {
      cerr << "Unable to connect a push supplier" << endl;
      return 1;
    }
    cerr << "Connection established" << endl;
    (new omni_thread( impl_is_ready, NULL))->start();
    char buf[1024];
    while (cin >> buf) {
      char *sendbuf= buf;
      istrstream istr(buf);
      long len;
      CORBA::String_var tmpstr;
      CORBA::Any v;
      if (istr >> len) {
	if (len>0) {
	  tmpstr= CORBA::string_alloc( len);
	  memset( (char *) tmpstr, 'x', len);
	  tmpstr[len-1]= '\0';
	  sendbuf= (char *) tmpstr;
	}
	else {
	  PerfTests pfs(-len);
	  pfs.length(-len);
	  for (long xq=0; xq<-len; xq++) {
	    pfs[xq].a= 'a';
	    pfs[xq].b= 1234;
	    pfs[xq].c= 0.98765;
	    pfs[xq].d= (const char *) "x";
	  }
	  v <<= pfs;
	}
      }
      if (len>0) {
#ifdef OMNIORB
	CORBA::Any::from_string fs(sendbuf, 0);
	v <<= fs;
#elif defined(ORBIX)
	v <<= sendbuf;
#endif
      }
      ECMcast::Time t= currtime();
      ppushcons->push( v);
      cerr << "Data sent " << t << endl;
    }
    try {
      ppushcons->disconnect_push_consumer();
    }
    catch (CORBA::Exception &ex) {
      cerr << "Error disconnecting proxy " << ex << endl;
      return 1;
    }
    cerr << "Event Channel disconnected" << endl;
    Bootstrap::boa_ptr->dispose( &pushsupi);
  }
  else {
    // typed channel
    //
    CosTypedEventChannelAdmin::TypedEventChannel_var ec;
    
    try {
      ec= CosTypedEventChannelAdmin::TypedEventChannel::_narrow(
		 nsroot->resolve( Bootstrap::create_name(argv[1])));
    }
    catch (...) {}
    
    if (CORBA::is_nil(ec)) {
      cerr << "Could not find Event Channel" << endl;
      return 1;
    }
    CosTypedEventChannelAdmin::TypedProxyPushConsumer_var ppushcons;
    try {
      CosTypedEventChannelAdmin::TypedSupplierAdmin_var sa=
	ec->for_suppliers();
      if (CORBA::is_nil(sa)) cerr << "TypeSupplierAdmin is nil" << endl;
      cerr << "for_suppliers" << endl;
      ppushcons= sa->obtain_typed_push_consumer("");
    }
    catch (CORBA::SystemException &se) {
      cerr << "SystemException" << se << endl;
    }
    catch (...) {}

    if (CORBA::is_nil(ppushcons)) {
      cerr << "Could not obtain ProxyPushConsumer from EventChannel" << endl;
      return 1;
    }

    PushSupplier_I &pushsupi= *(new PushSupplier_I());
    CONNECT_IMPL(Bootstrap::boa_ptr,&pushsupi);
    try {
      ppushcons->connect_push_supplier( &pushsupi);
    }
    catch (CosEventChannelAdmin::AlreadyConnected &) {
      cerr << "Already connected" << endl;
      return 1;
    }
    catch (...) {
      cerr << "Unable to connect a push supplier" << endl;
      return 1;
    }
    cerr << "Connection established" << endl;
    CORBA::Object_var objconsumer;
    try {
      objconsumer= ppushcons->get_typed_consumer();
    }
    catch (...) {
      cerr << "Unable to get a typed consumer" << endl;
      return 1;
    }
    TestIntf_var ti;
    try {
      ti= TestIntf::_narrow(objconsumer);
    }
    catch (...) {}
    if (CORBA::is_nil(ti)) {
      cerr << "Unable to narrow a typed consumer to a final interface" << endl;
      return 1;
    }
    cerr << "Got typed consumer " << endl;
    char buf[1024];
    while (cin >> buf) {
      try {
	char *sendbuf= buf;
	istrstream istr(buf);
	long len;
	CORBA::String_var tmpstr;
	if (istr >> len) {
	  if (len>0) {
	    tmpstr= CORBA::string_alloc( len);
	    memset( (char *) tmpstr, 'x', len);
	    tmpstr[len-1]= '\0';
	    sendbuf= (char *) tmpstr;
	    ECMcast::Time t= currtime();
	    cerr << "Sending data" << t << endl;
	    ti->operString(tmpstr);
	    continue;
	  }
	  else {
	    PerfTests pfs(-len);
	    pfs.length(-len);
	    for (long xq=0; xq<-len; xq++) {
	      pfs[xq].a= 'a';
	      pfs[xq].b= 1234;
	      pfs[xq].c= 0.98765;
	      pfs[xq].d= (const char *) "x";
	    }

	    ECMcast::Time t= currtime();
	    ti->operPerf(pfs);
	    cerr << "Data sent " << t << endl;
	  }
	}

	if (!strcmp(buf,"1")) {
	  // operBoolean
	  ti->operBoolean( 1, 3.14159);
	}
	else if (!strcmp(buf,"2")) {
	  // operDouble
	  ti->operDouble( 3.14159, 1);
	}
	else if (!strcmp(buf,"3")) {
	  // operString
	  ti->operString("ala ma psa");
	}
	else if (!strcmp(buf,"4")) {
	  // operAny
	  CORBA::Any v;
	  v <<= (CORBA::Double) 2.72;
	  ti->operAny( v);
	}
	else {
	  CORBA::Any v;
#ifdef OMNIORB
	  CORBA::Any::from_string fs(buf, 0);
	  v <<= fs;
#elif defined(ORBIX)
	  v <<= buf;
#endif
	  ppushcons->push( v);
	}
      }
      catch (CORBA::Exception &ex) {
	cerr << ex << endl;
	return 1;
      }
    }
    try {
      ppushcons->disconnect_push_consumer();
    }
    catch (CORBA::Exception &ex) {
      cerr << "Error disconnecting proxy " << ex << endl;
      return 1;
    }
    cerr << "Event Channel disconnected" << endl;
    Bootstrap::boa_ptr->dispose( &pushsupi);    
  }
  return 0;
}

