#ifndef __TIMESTAMP_HH__
#define __TIMESTAMP_HH__

class TimeStampFactory;

class TimeStamp {
public:
  long sec;
  long nsec;

  TimeStamp( TimeStampFactory &);
  TimeStamp() :
    sec(0), nsec(0) {}
  TimeStamp( long _sec, long _nsec) :
    sec(_sec), nsec(_nsec) {}

  TimeStamp& normalize() {
    while (nsec>=1000000000L) {
      nsec-= 1000000000L;
      ++sec;
    }
    return *this;
  }

  int operator < ( const TimeStamp &df) const {
    TimeStamp t1( *this);
    t1.normalize();
    TimeStamp t2( df);
    t2.normalize();
    return (t1.sec<t2.sec) ? 1 : 
      ((t1.sec>t2.sec) ? 0 :
       ((t1.nsec<t2.nsec) ? 1 : 0));
  }

  TimeStamp diff ( const TimeStamp &sb) const {
    TimeStamp t1(*this);
    TimeStamp t2(sb);
    if (*this < sb) {
      t1= sb;
      t2= *this;
    }
    t1.normalize();
    t2.normalize();
    TimeStamp ret;
    if (t1.nsec<t2.nsec) {
      ret.nsec= (t1.nsec+1000000000L)-t2.nsec;
      ++(t2.sec);
    }
    else ret.nsec= t1.nsec-t2.nsec;
    ret.sec= t1.sec-t2.sec;
    return ret;
  }  
};

class TimeStampFactory {
public:
  virtual TimeStamp instantiate()= 0;
};

extern TimeStampFactory *default_tsf;


#endif // __TIMESTAMP_HH__

//
