// $Id$
//
// $Source$
//
// $Log$
//

#ifndef NO_MCAST

#include "mtp.hh"
#include "mtpsource.hh"
#include "mtpsink.hh"
#include "mtpwebd.hh"
#include <iostream.h>
#include <string.h>
#include <stdio.h>

#include "mcast.hh"

#ifndef __WIN32__
#include <unistd.h>
#endif

//--------------------------------------------------------------MTPMCFactory--
//
class MTPMCFactory : public McastFactory {
public:
  MTPMCFactory() {}

  virtual void init( int & argc, char ** &argv);
  virtual McastSourceFactory *connect_as_master();
  virtual McastSinkFactory *connect_as_slave();
  virtual ~MTPMCFactory();

  friend class MTPMCSourceFactory;
  friend class MTPMCSinkFactory;
};

McastFactory *default_mcast_factory= new MTPMCFactory();

class MTPMCSource;
class MTPMCSink;
//--------------------------------------------------------MTPMCSourceFactory--
//
class MTPMCSourceFactory : public McastSourceFactory {
  XSTD::map<PacketMgr*,MTPMCSource*,XSTD::less<PacketMgr*> > mgr2source;
  omni_mutex guard;
public:
  MTPMCSourceFactory();
  virtual McastSource *create( McastSlaveConnectedCallback *cback,
			       int new_channel_hint);
  virtual ~MTPMCSourceFactory();
};

//----------------------------------------------------------MTPMCSinkFactory--
//
class MTPMCSinkFactory : public McastSinkFactory {
  XSTD::map<PacketMgr*,MTPMCSink*,XSTD::less<PacketMgr*> > mgr2sink;
  omni_mutex guard;
public:
  MTPMCSinkFactory();
  virtual McastSink *create( const char *unique_description);
  virtual ~MTPMCSinkFactory();
};

//---------------------------------------------------------------MTPMCSource--
//
class MTPMCSource : public McastSource {
  MTPSource &src;
  omni_mutex guard;
public:
  MTPMCSource( MTPSource &_src);

  virtual char *get_unique_description();

  virtual void start_message();
  virtual void end_message();

  virtual void write( void *data, unsigned long len);
  virtual ~MTPMCSource();
};

//-----------------------------------------------------------------MTPMCSink--
//
class MTPMCSink : public McastSink {
  MTPSink &sink;
public:
  MTPMCSink( MTPSink &_sink);
  
  virtual void read( void *buf, unsigned long buflen);
  virtual ~MTPMCSink();
};

//--------------------------------------------------------------MTPMCFactory--
//
void MTPMCFactory::
init( int & argc, char ** &argv) {
  PacketMgrFactory::init( argc, argv);
}

McastSourceFactory * MTPMCFactory::
connect_as_master() {
  return new MTPMCSourceFactory();
}

McastSinkFactory * MTPMCFactory::
connect_as_slave() {
  return new MTPMCSinkFactory();
}

MTPMCFactory::
~MTPMCFactory() {
}

//--------------------------------------------------------MTPMCSourceFactory--
//
static GeneralLock mylock;

MTPMCSourceFactory::
MTPMCSourceFactory() {  
}

McastSource * MTPMCSourceFactory::
create( McastSlaveConnectedCallback *_cback,
	int new_channel_hint) {

  omni_mutex_lock oml(guard);

  PacketMgr *pm( packetmgrfactory.assign_channel());
  if (mgr2source.find( pm) == mgr2source.end()) {
    init_mtp_deamon( *pm);
    MTPSource *src;
    src= new MTPSource( *pm);
    src->run();
#ifdef __WIN32__
    Sleep(1);
#else
    sleep(1);
#endif
    return (mgr2source[pm]= new MTPMCSource( *src));
  }
  else return mgr2source[pm];
};

MTPMCSourceFactory::
~MTPMCSourceFactory() {
}

//----------------------------------------------------------MTPMCSinkFactory--
//
MTPMCSinkFactory::
MTPMCSinkFactory() {  
}

McastSink *MTPMCSinkFactory::
create( const char *unique_description) {
  omni_mutex_lock oml(guard);

  PacketMgr *pm( packetmgrfactory.find_or_create( unique_description));
  if (mgr2sink.find( pm) == mgr2sink.end()) {
    MTPSink *sink;
    sink= new MTPSink( *pm);
    return (mgr2sink[pm]= new MTPMCSink( *sink));
  }
  else return mgr2sink[pm];
}

MTPMCSinkFactory::
~MTPMCSinkFactory() {
}

//---------------------------------------------------------------MTPMCSource--
//
MTPMCSource::
MTPMCSource( MTPSource &_src) :
  McastSource( &mylock),
  src( _src) 
{
}

char * MTPMCSource::
get_unique_description() {
  char *ret= new char[100];
  sprintf(ret, "%s %u",
	  src.get_packetmgr().get_address(),
	  src.get_packetmgr().get_listen_port());
  //  cerr << "get_unique_description " << ret << endl;
  return ret;
}

void MTPMCSource::
start_message() {
  src.start_message();
}

void MTPMCSource::
end_message() {
  src.end_message();
}

void MTPMCSource::
write( void *data, unsigned long len) {
  src.write( data, len);
}

MTPMCSource::
~MTPMCSource() {
  delete &src;
}

//-----------------------------------------------------------------MTPMCSink--
//

MTPMCSink::
MTPMCSink( MTPSink &_sink) :
  sink( _sink) 
{
}
  
void MTPMCSink::
read( void *buf, unsigned long buflen) {
  sink.read(buf, buflen);
}

MTPMCSink::
~MTPMCSink() {
  delete &sink;
}

#endif // NO_MCAST

// end of file $File$
//
