// $Id: mtpsink.cc,v 1.1 1998/11/11 02:22:30 mencda Exp mencda $
//
// $Source: /home/mencda/thesis/mtp/mtpsink.cc,v $
//
// $Log: mtpsink.cc,v $
// Revision 1.1  1998/11/11 02:22:30  mencda
// Initial revision
//
//
//

#ifndef NO_MCAST

#include "mtpsink.hh"
#include <iostream.h>

MTPSink::~MTPSink() {
  packetmgr.unregister_handler( *this, my_ucast_id);  
}

void MTPSink::init() {
  my_ucast_id= mtpidmgr.assign_id();
  trace.start(1000) << "my_ucast_id " << my_ucast_id << '\n';
  web= MTPWeb( 300, // msec -heartbeat 
	       15,  // packets -window
	       10,  // retention
	       MAX_MTP_PACKET_SIZE, my_ucast_id, packetmgr.get_ucast_port());
  MTPPacket tryjoin;

  omni_mutex_lock oml(guard);
  packetmgr.register_handler( *this, my_ucast_id, my_ucast_id);

  for (int try_client=0; try_client<10; try_client++) {
    tryjoin.create_join_request( MTPHeader::MEMBER_CONSUMER,
				 web);
    packetmgr.send_mcast( tryjoin);
    unsigned long sec,nsec;
    omni_thread::get_time( &sec,&nsec, 0, 30000000 /* 30 msec */);
    cond_wait.timedwait( sec, nsec);
    if (state != STATE_INIT) break;
  }
  if (state == STATE_INIT) {
    trace.start(1000) << "Failed to find web master" << '\n';
    state= STATE_FAILED;
    packetmgr.unregister_handler( *this, my_ucast_id);
    cerr << "NoMasterActive" << endl;
    throw McastFactory::NoMasterActive();
  }
  trace.start(1000) << "MTPSink initialized" << '\n';
}

void MTPSink::send_naks( MTPPacket &packet) {
  // first approach: send naks for all missing packets
  //
  trace.start(10000) << "send_naks" << '\n';

  XSTD::vector<NAKRange> ranges;
  NAKRange nakr;

  u48 msg_no_first= last_message_accepted+(u32) 1;
  u16 pck_no_first= assemble[msg_no_first].last_accepted+1;
  for (Msgs::iterator it= assemble.begin();
       it != assemble.end();
       it++) {

    MessageAssemble &mass= (*it).second;
    for (MessageAssemble::MapAssemble::iterator xit= mass.packets.begin();
	 xit != mass.packets.end();
	 xit++) {
      if ((*xit).second.header.msg_seq_number == msg_no_first &&
	  (*xit).second.header.packet_seq_number == pck_no_first) {
	if ((*xit).second.header.packet_type == MTPHeader::T_DATA &&
	    (*xit).second.header.packet_modifier == MTPHeader::T_DATA_EOM) {
	  msg_no_first= msg_no_first+(u32) 1;
	  pck_no_first= 1;
	}
	else 
	  pck_no_first++;
	continue;
      }
      nakr.min_message= msg_no_first;
      nakr.min_packet= pck_no_first;
      nakr.max_message= (*xit).second.header.msg_seq_number;
      nakr.max_packet= (*xit).second.header.packet_seq_number;
      ranges.push_back( nakr);
      trace.start(10000) 
	<< "NAK [ " << nakr.min_message << ", "
	<< nakr.min_packet << "] -> [ " << nakr.max_message << ", "
	<< nakr.max_packet << "]" << '\n';
      msg_no_first= (*xit).second.header.msg_seq_number;
      pck_no_first= (*xit).second.header.packet_seq_number+1;
      if ((*xit).second.header.packet_type == MTPHeader::T_DATA &&
	  (*xit).second.header.packet_modifier == MTPHeader::T_DATA_EOM) {
	msg_no_first= msg_no_first+(u32) 1;
	pck_no_first= 1;
      }
    }
  }
  if ((msg_no_first < packet.header.msg_seq_number ||
       msg_no_first == packet.header.msg_seq_number &&
       pck_no_first < packet.header.packet_seq_number) &&
      packet.header.packet_type == MTPHeader::T_EMPTY &&
      packet.header.packet_modifier == MTPHeader::T_EMPTY_DALLY) {

    nakr.min_message= msg_no_first;
    nakr.min_packet= pck_no_first;
    nakr.max_message= packet.header.msg_seq_number;
    nakr.max_packet= packet.header.packet_seq_number;
    ranges.push_back( nakr);
    trace.start(10000) 
      << "NAK [ " << nakr.min_message << ", "
      << nakr.min_packet << "] -> [ " << nakr.max_message << ", "
      << nakr.max_packet << "]" << '\n';    
  }

  out_packet.create_nak_request( web, producer.id, ranges);
  out_packet.addrfrom= producer.addr;
  out_packet.addrfrom_len= producer.addr_len;
  packetmgr.send_ucast_reply( out_packet, packet.header.reply_port);
  trace.start(1000) << "send_naks...done" << '\n';
}

void MTPSink::handle_data_packet( MTPPacket &packet, int dally) {
  if (!producer.id && !dally) {
    producer.addr= packet.addrfrom;
    producer.addr_len= packet.addrfrom_len;
    producer.id= packet.header.src_id;
  }

  if (packet.header.msg_seq_number <= last_message_accepted) {
    trace.start(10000)
      << "last_message_accepted " 
      << last_message_accepted
      << "  msg_seq_number " 
      << packet.header.msg_seq_number 
      << "MTPSink  ; retransmission" << '\n';
    // a retransmission; we have completed assembly of these
    // messages, just ignore them
    return;
  }
  MessageAssemble &ma= assemble[packet.header.msg_seq_number];
   
  if (ma.accepted) {
    trace.start(10000) 
      << "MTPSink::data_packet ; retransmission ; already accepted" << '\n';
    // a retransmission; we have completed assembly of this message
    return;
  }

  u16 last_packet_no= ma.last_accepted;
  u16 next_packet_no= last_packet_no+1;
  trace.start(10000)
    << "next_packet_no " << next_packet_no << '\n'
    << "packet_seq_no " << packet.header.packet_seq_number 
    << '\n';
  trace.start(10000) 
    << "message (last, current) (" << last_message_accepted << ", "
    << packet.header.msg_seq_number << ")" << '\n';
  if (packet.header.packet_seq_number == next_packet_no &&
      last_message_accepted+1 == packet.header.msg_seq_number) {
    if (dally) {
      // it's really a fake data message advertising new 
      // message sequence number
      return;
    }
    //    cerr << "it is safe to move accepted packets to the buffer" 
    //	 << endl;
    // it is safe to move accepted packets to the buffer
    trace.start(10000)
      << "Retiring " << packet.header.msg_seq_number 
      << '.' << packet.header.packet_seq_number << '\n';
    grow_buffer.push_front( MessageReady());
    MessageReady &mr= grow_buffer.front();
    mr.prepare( packet.data_length);
    memcpy( mr.buf, packet.data, packet.data_length);
    MTPHeader header_accepted( packet.header);
    ma.last_accepted= next_packet_no;
    MessageAssemble::MapAssemble::iterator it( ma.packets.begin());
    for (; it != ma.packets.end(); ) {
      if (ma.last_accepted+1 != (*it).second.header.packet_seq_number) {
	cond_wait_for_data.signal();
	return;
      }
      MTPPacket &pck= (*it).second;
      grow_buffer.push_front( MessageReady());
      MessageReady &nmr= grow_buffer.front();
      nmr.prepare( pck.data_length);
      memcpy( nmr.buf, pck.data, pck.data_length);
      header_accepted= pck.header;
      trace.start(10000)
	<< "Retiring " << pck.header.msg_seq_number 
	<< '.' << pck.header.packet_seq_number << '\n';
      ma.last_accepted++;
      MessageAssemble::MapAssemble::iterator erit( it);
      it++;
      ma.packets.erase( erit);
    }
    if (it == ma.packets.end()) {
      //      cerr << "it == ma.packets.end()" << endl;
      // all packets in this message has been moved to the buffer
      if (header_accepted.packet_type == MTPHeader::T_DATA &&
	  header_accepted.packet_modifier == MTPHeader::T_DATA_EOM ||
	  header_accepted.packet_type == MTPHeader::T_EMPTY &&
	  header_accepted.packet_modifier == MTPHeader::T_EMPTY_CANCEL) {
	//	cerr << "end of message packet accepted" << endl;
	//	cerr << "checking for " << last_message_accepted+1 << endl;
	// end of message packet accepted
	assemble.erase( last_message_accepted= header_accepted.msg_seq_number);
	MessageAssemble *mai;
	while ( (mai= &(assemble[last_message_accepted+1]))->accepted) {
	  for (MessageAssemble::MapAssemble::iterator 
		 xit= mai->packets.begin();
	       xit != mai->packets.end();
	       xit++) {
	    MTPPacket &pck= (*xit).second;
	    trace.start(10000)
	      << "Retiring " << pck.header.msg_seq_number 
	      << '.' << pck.header.packet_seq_number << '\n';
	    grow_buffer.push_front( MessageReady());
	    MessageReady &xmr= grow_buffer.front();
	    xmr.prepare( pck.data_length);
	    memcpy( xmr.buf, pck.data, pck.data_length);
	  }
	  last_message_accepted= last_message_accepted+(u32) 1;
	  assemble.erase( last_message_accepted);
	}
	mai= &(assemble[last_message_accepted+1]);
	if (!mai->accepted) {
	  u16 x= mai->last_accepted+1;
	  for (MessageAssemble::MapAssemble::iterator 
		 xit( mai->packets.begin());
	       xit != mai->packets.end(); ) {
	    if (x++ != (*xit).second.header.packet_seq_number) break;
	    MTPPacket &pck= (*xit).second;
	    trace.start(10000)
	      << "Retiring " << pck.header.msg_seq_number 
	      << '.' << pck.header.packet_seq_number << '\n';
	    grow_buffer.push_front( MessageReady());
	    MessageReady &xmr= grow_buffer.front();
	    xmr.prepare( pck.data_length);
	    memcpy( xmr.buf, pck.data, pck.data_length);	    
	    mai->last_accepted++;
	    MessageAssemble::MapAssemble::iterator erit( xit);
	    xit++;
	    mai->packets.erase( erit);
	  }
	}
      }
    }
    cond_wait_for_data.signal();
  }
  else {
    if (ma.last_accepted >= packet.header.packet_seq_number) {
      // a retransmission
      return;
    }

    // we have a hole in message assembly sequence
    // this message has not been accepted yet
    //

    if (!dally) {
      ma.packets[ packet.header.packet_seq_number]= packet; 
    
      u16 x= ma.last_accepted+1;
      for (MessageAssemble::MapAssemble::iterator it( ma.packets.begin());
	   it != ma.packets.end(); 
	   it++) {
	if (x++ != (*it).second.header.packet_seq_number) break;
	if ((*it).second.header.packet_type == MTPHeader::T_DATA &&
	    (*it).second.header.packet_modifier == MTPHeader::T_DATA_EOM) {
	  //	  cerr << "message " << packet.header.msg_seq_number
	  //	       << " accepted" << endl;
	  ma.accepted= 1;
	}
      }
    }
    send_naks( packet);
  }
}

void MTPSink::handle( MTPPacket &packet) {
  omni_mutex_lock oml(guard);

  static int err= 0;
  //  if (rand() < RAND_MAX/5) {
  //    trace.start(1000) << "Simulate lost packet" << '\n';
  //    return;
  //  }
  u16 reply_port;
  reply_port= packet.header.reply_port;

  switch (packet.header.packet_type) {
  case MTPHeader::T_JOIN:
    switch (packet.header.packet_modifier) {
    case MTPHeader::T_JOIN_CONFIRM:
      if (state == STATE_INIT) {
	web.update_after_join( packet);
	master.addr= packet.addrfrom;
	master.addr_len= packet.addrfrom_len;
	master.reply_port= packet.header.reply_port;
	last_message_accepted= packet.header.msg_seq_number;
	packetmgr.register_handler( *this, 
				    web.mcast_connection_id(),
				    my_ucast_id);
	state= STATE_ESTABLISHED;
      }
      break;
    case MTPHeader::T_JOIN_DENY:
      if (state == STATE_INIT) state= STATE_FAILED;
      break;
    }
  case MTPHeader::T_QUIT:
    // implement: quitting a web on demand
    break;
  case MTPHeader::T_IS_MEMBER:
    switch (packet.header.packet_modifier) {
    case MTPHeader::T_IS_MEMBER_REQUEST:
      // all I can do is to confirm the membership of myself
      if (packet.header.dst_id == my_ucast_id)
	packet.create_is_member_confirm( web);
      else
	packet.create_is_member_deny( web);
      packetmgr.send_ucast_reply( packet, reply_port);
      break;
    }
    break;
  case MTPHeader::T_EMPTY:
    switch (packet.header.packet_modifier) {
    case MTPHeader::T_EMPTY_DALLY:
      trace.start(10000)
	<< "Data! DALLY [ " << packet.header.msg_seq_number << ", "
	<< packet.header.packet_seq_number << "]\n";
      handle_data_packet( packet, 1);
      break;
    case MTPHeader::T_EMPTY_CANCEL:
      break;
    case MTPHeader::T_EMPTY_HIBERNATE:
      break;
    }    
    break;
  case MTPHeader::T_NAK:
    if (packet.header.packet_modifier == MTPHeader::T_NAK_DENY) {
      trace.start(10000) << "NAK_DENY" << '\n';
      state= STATE_FAILED;
      cond_wait_for_data.signal();
    }
    break;
  case MTPHeader::T_DATA:
    trace.start(10000)
      << "Data! [ " << packet.header.msg_seq_number << ", "
      << packet.header.packet_seq_number << "]\n";
    handle_data_packet( packet, 0);
    break;
  }
}

void MTPSink::read( void *buf, unsigned long buflen) {
  omni_mutex_lock oml(guard);
  unsigned char *xbuf= (unsigned char *) buf;

  while (buflen>0) {
    while (grow_buffer.empty()) cond_wait_for_data.wait();
    if (state == STATE_FAILED) throw McastSink::LostData();
    MessageReady &mr= grow_buffer.back();
    u16 avail_len= mr.len-mr.rd_ptr;
    u16 len= (buflen>avail_len) ? avail_len : buflen;
    memcpy( xbuf, &mr.buf[mr.rd_ptr], len);
    trace.start(900) << "read " << len << '\n';
    xbuf+= len;
    mr.rd_ptr+= len;
    buflen-= len;
    if (mr.rd_ptr == mr.len) {
      //      cerr << "Message consumed" << endl;
      grow_buffer.pop_back();
    }
  }
}


#endif // NO_MCAST

//
